// Simple test for PIC16F887
// amateur code. Author accepts no responsibility for anything, ever
// read NEC IR remote code
// connect IR receiver to RB0
// 8MHz internal oscillator
// PORTC is used for debug/diagnostic LEDs
// PORTD used for 16x2 LCD
// code is messy, but it is an attempt to fix the terrible code I was given - should have started from scratch :(

// Current progress 16 August 2019
// 1 red remote works better than Logik TV remote

/*----------------------------
Extended NEC protocol: 1/4 or 1/3 duty cycle 38khz (21 pulses per bit approx)
9ms mark then 4.5ms space	header
560us mark, 560us space		zero
560us mark, 1.69ms space	one
4 bytes, LSB first
560us mark - long space
stop bit of one
option 1 = repeat every 110ms (if wanted), but just 9ms mark, 2.25 space, 560us mark (no data)
option 2 = repeat full message
00ff = default address, custom codes available, byte 4 is inverse of byte 3
byte1 is low byte of address, byte2 is either inverse of byte 1 or high byte
---------------------------------*/

sbit LCD_RS at RD1_bit;
sbit LCD_EN at RD2_bit;
sbit LCD_D4 at RD4_bit;
sbit LCD_D5 at RD5_bit;
sbit LCD_D6 at RD6_bit;
sbit LCD_D7 at RD7_bit;
sbit LCD_RS_Direction at TRISD1_bit;
sbit LCD_EN_Direction at TRISD2_bit;
sbit LCD_D4_Direction at TRISD4_bit;
sbit LCD_D5_Direction at TRISD5_bit;
sbit LCD_D6_Direction at TRISD6_bit;
sbit LCD_D7_Direction at TRISD7_bit;
// End LCD module connections

bit nec_ok;                                                // indicate NEC message is ready
char text[5];                                              // general purpose for conversions
unsigned short nec_state;                                  // count through decoding stages
unsigned short command;                                    // 8 bit NEC command
unsigned short inv_command;                                // 8 bit inverse of command
unsigned short bit_count;                                  // bit counter for IR bits
unsigned short addressh;                                   // high byte of NEC address
unsigned short addressl;                                   // low byte of NEC address
unsigned int address;                                      // 16 bit NEC address
unsigned int timer_value;                                  // store current timer 1 value for current bit
unsigned long nec_code;                                    // 32 bit complete NEC message

void Interrupt() {
    if (INTF_bit && INTE_bit){                             // PORTB0 external interrupt
        INTF_bit = 0;                                      // reset interrupt
        if(nec_state != 0){
            timer_value = (TMR1H << 8) + TMR1L;            // Store Timer1 value
            TMR1H = 0;                                     // Reset Timer1
            TMR1L = 0;
        }
        switch(nec_state){
          case 0 :                                         // Start receiving IR data (we're at the beginning of 9ms pulse)
            TMR1H = 0;                                     // Reset Timer1
            TMR1L = 0;
            TMR1ON_bit = 1;                                // Enable Timer1
            nec_state = 1;                                 // Next state: end of 9ms pulse (start of 4.5ms space)
            bit_count = 0;
            PORTC = 0;
            INTEDG_bit = 1;                                // next Interrupt rising edge
            break;
          case 1 :                                         // End of 9ms pulse
            //PORTC = 1;
            if((timer_value > 9500) || (timer_value < 8500)){         // Invalid interval ==> stop decoding and reset
                nec_state = 0;                             // Reset decoding process
                TMR1ON_bit = 0;                            // Disable Timer1
            }
            else{
                nec_state = 2;                             // Next state: end of 4.5ms space (start of 562s pulse)
                INTEDG_bit = 0;                            // next Interrupt falling edge
            }
            break;
          case 2 :                                         // End of 4.5ms space
            nec_code = 0;                                  // clear out all data bits when header found
            if((timer_value > 5000) || (timer_value < 4000)){
              nec_state = 0;                               // Reset decoding process
              TMR1ON_bit = 0;                              // Disable Timer1
            }
            else{
                nec_state = 3;                             // Next state: end of 562s pulse (start of 562s or 1687s space)
                INTEDG_bit = 1;                            // next Interrupt rising edge
            }
            break;
          case 3 :                                         // End of 562s pulse
            if((timer_value > 700) || (timer_value < 400)){           // Invalid interval ==> stop decoding and reset
                TMR1ON_bit = 0;                            // Disable Timer1
                nec_state = 0;                             // Reset decoding process
            }
            else{
                nec_state = 4;                             // Next state: end of 562s or 1687s space
                INTEDG_bit = 0;                            // next Interrupt falling edge
            }
            break;
          case 4 :                                         // End of 562s or 1687s space
            if((timer_value > 2800) || (timer_value < 400)){           // Time interval invalid ==> stop decoding
                TMR1ON_bit = 0;                            // Disable Timer1
                nec_state = 0;                             // Reset decoding process
                break;
            }
            if( timer_value > 800){                       // If space width > 0.6ms (short space)
                nec_code |= 1UL << (bit_count);          // Write 1 to bit (31 - i)
            }
            else{                                           // If space width 1.8ms (long space)
                //nec_code |= 0 << (31 - bit_count);          // Write 0 to bit (31 - i)
            }
            bit_count++;
            if(bit_count > 31){                            // If all bits are received
                nec_ok = 1;                                // Decoding process OK
                INTE_bit = 0;                              // Disable PORTB external interrupt
                return;
            }
            nec_state = 3;                                 // Next state: end of 562s pulse (start of 562s or 1687s space)
            INTEDG_bit = 1;                                // next Interrupt rising edge
            break;
        }
    }
    if (TMR1IF_bit){                                       // Timer1 ISR - time-out after 65 milliseconds
        TMR1IF_bit = 0;                                    // Clear Timer1 overflow flag bit
        nec_state = 0;                                     // Reset decoding process
        TMR1ON_bit = 0;                                    // Disable Timer1
    }
}

void main() {
    OSCCON = 0X70;                                         // 8 MHz internal oscillator
    TRISB = 0x01;                                          // RB0 input for IR receiver, rest output
    TRISD = 0x00;                                          // all output
    ANSEL = 0x00;                                          // all digital
    ANSELH = 0;                                            // Configure all PORTB pins as digital
    nec_ok = 0;                                            // initialise
    nec_state = 0;                                         // initial state
    Lcd_Init();                                            // Initialize LCD module
    Lcd_Cmd(_LCD_CURSOR_OFF);                              // cursor off
    Lcd_Cmd(_LCD_CLEAR);                                   // clear LCD
    TMR1IF_bit = 0;                                        // Clear Timer1 overflow interrupt flag bit
    TMR1IE_bit = 1;                                        // Enable Timer1 overflow interrupt
    T1CON = 0x10;                                          // Set Timer1 clock source to internal with 1:2 prescaler (Timer1 clock = 1MHz)
    INTCON = 0x80;                                         // global Interrupt enable
    OPTION_REG = 0x80;                                     // pull-ups disabled, ext INT falling edge
    Lcd_Out(1,1,"NEC decode IR");
    Lcd_Out(2,1, "16 Aug 2019");
    Delay_ms(2000);                                        // allow time to see first message
    Lcd_Cmd(_LCD_CLEAR);
    Lcd_Out(1, 1, "Address:0x0000");
    Lcd_Out(2, 1, "Com:0x00 In:0x00");
    Delay_ms(1000);
    INTE_bit = 1;                                          // Enable PORTB external Interrupt
    INTF_bit = 0;                                          // clear possible old interrupt

    while(1) {
        while(!nec_ok);                                    // Wait until NEC code received
        nec_ok = 0;                                        // Reset decoding process
        nec_state = 0;
        TMR1ON_bit = 0;                                    // Disable Timer1
        addressh = nec_code;                               // swap address bytes around ...
        addressl = nec_code >> 8;
        address = (addressh << 8) + addressl;              // ... then re-combine
        command = nec_code >> 16;
        inv_command = nec_code >> 24;
        IntToHex(address, text);                           // Save address in string text with hex format
        Lcd_Out(1, 11, text);                              // Display address
        ByteToHex(command, text);                          // Save command in string text with hex format
        Lcd_Out(2, 7, text);                               // Display command
        ByteToHex(inv_command, text);                      // Save inverted command in string text with hex format
        Lcd_Out(2, 15, text);                              // Display inverted command
        INTE_bit = 1;                                      // Enable PORTB external interrupt
    }
}